use std::{path::PathBuf, time::Instant, net::Ipv4Addr, fs::File, io::{Write, BufReader, BufRead}};

use anyhow::anyhow;
use clap::Parser;
use log::info;
use responder_common::bloom_filter;

#[derive(Debug, Parser)]
pub struct Options {
    pub ip_path: PathBuf,
    pub out_path: Option<PathBuf>,
    #[clap(short, long)]
    pub force: bool,
}

pub fn build_filter(opts: Options) -> Result<(), anyhow::Error> {
    let out_path = opts.out_path.unwrap_or_else(|| {
        let mut ext = opts.ip_path.extension().unwrap_or_default().to_os_string();
        ext.push(".bfb");
        opts.ip_path.with_extension(ext)
    });

    if out_path.exists() {
        if !opts.force {
            return Err(anyhow!("Output path {} already exists, use --force/-f to overwrite", out_path.display()));
        } else if !out_path.is_file() {
            return Err(anyhow!("Can not force overwrite output path {}: is not a file", out_path.display()));
        }
    }

    let start = Instant::now();

    // let mut addresses = Vec::new();
    let mut bloom_filter: Vec<u8> = Vec::with_capacity(bloom_filter::BITS>>3);
    bloom_filter.resize(bloom_filter::BITS >> 3, 0);

    let f = File::open(opts.ip_path)?;
    let reader = BufReader::new(f);
    let mut clash_count = 0;
    for (i, line) in reader.lines().enumerate() {
        if i % 1_000_000 == 0{
            info!("processed {}M ips; {} clashes",i / 1_000_000, clash_count);
        }
        let line = line?;
        let key = u32::from(line.parse::<Ipv4Addr>()?);
        // addresses.push(key);
        let mut clashes = true;
        for i in 0..bloom_filter::HASH_COUNT {
            let hash = bloom_filter::hash(key, i);
            let i = hash >> 3;
            let byte_i = hash & 7;
            // info!("{} {} {} {}", line, hash, i, byte_i);
            clashes = clashes && ((bloom_filter[i as usize] >> (7 - byte_i)) & 1 == 1);
            bloom_filter[i as usize] |= 0x80 >> byte_i;
        }
        if clashes {
            clash_count += 1;
        }
    }
    info!("clash count: {}", clash_count);

    let filter = start.elapsed();

    let mut file = File::create(out_path)?;
    let byte_slice: &[u8] = bloom_filter.as_slice();
    file.write_all(byte_slice)?;

    let end = start.elapsed();

    info!("build filter: {:?} write: {:?} | total: {:?}", filter, end-filter, end);
    Ok(())
}
